import warnings
from datetime import datetime
from datetime import timezone as dtimezone
from typing import TYPE_CHECKING, Any, Dict, Optional

from apscheduler.util import undefined
from esmerald.conf import settings
from esmerald.exceptions import ImproperlyConfigured
from esmerald.schedulers.apscheduler.types import TriggerType
from esmerald.types import SchedulerType
from esmerald.utils.module_loading import import_string

if TYPE_CHECKING:
    from esmerald.applications import Esmerald
    from pydantic.typing import AnyCallable


class Scheduler:
    def __init__(
        self,
        app: Optional["Esmerald"] = None,
        scheduler_class: Optional["SchedulerType"] = None,
        tasks: Optional[Dict[str, str]] = None,
        timezone: Optional[dtimezone] = None,
        configurations: Optional[Dict[str, str]] = None,
    ) -> None:
        self.app = app
        self.tasks = tasks or {}
        self.timezone = timezone
        self.scheduler_class = scheduler_class

        if not self.scheduler_class and self.app.enable_scheduler:
            raise ImproperlyConfigured(
                "It cannot start the scheduler if there is not scheduler_class declared."
            )

        for (
            task,
            module,
        ) in self.tasks.items():
            if not isinstance(task, str) or not isinstance(module, str):
                raise ImproperlyConfigured("The dict of tasks must be Dict[str, str].")

        self.handler = self.get_scheduler(scheduler=self.scheduler_class, timezone=self.timezone)

        if not self.tasks:
            warnings.warn(
                "Esmerald is starting the scheduler, yet there are " "no tasks declared.",
                UserWarning,
                stacklevel=2,
            )

        self.register_events(app=self.app)
        self.register_tasks(tasks=tasks)

    def register_tasks(self, tasks: Dict[str, str]):
        """
        Registers the tasks into the Schedulers
        """
        for task, _module in tasks.items():
            imported_task = f"{_module}.{task}"
            scheduled_task: "Task" = import_string(imported_task)

            if not scheduled_task.is_enabled:
                continue

            try:
                scheduled_task.add_job(self.handler)
            except Exception as e:
                raise ImproperlyConfigured(str(e))

    def register_events(self, app: "Esmerald") -> None:
        """
        Registers the scheduler start/stop events in the application.
        """

        @app.on_event("startup")
        def start_scheduler() -> None:
            self.handler.start()

        @app.on_event("shutdown")
        def stop_scheduler() -> None:
            self.handler.shutdown()

    def get_scheduler(
        self, scheduler: "SchedulerType", timezone: Optional[dtimezone] = None
    ) -> SchedulerType:
        """
        Initiates the scheduler from the given time.
        If no value is provided, it will default to AsyncIOScheduler.

        The value of `scheduler_class` can be overwritten by any custom settings.

        Args:
            scheduler: AsyncIOScheduler

        Return:
            None
        """
        if not timezone:
            timezone = settings.timezone

        return scheduler(timezone=timezone)


class Task:
    """
    Base for the scheduler decorator that will auto discover the
    tasks in the application and add them to the internal scheduler.
    """

    def __init__(
        self,
        *,
        name: Optional[str] = None,
        trigger: Optional[TriggerType] = None,
        identifier: Optional[str] = None,
        misfire_grace_time: Optional[int] = None,
        coalesce: Optional[bool] = None,
        max_intances: Optional[int] = None,
        next_run_time: Optional[datetime] = None,
        jobstore: Optional[str] = "default",
        executor: Optional[str] = "default",
        replace_existing: bool = False,
        args: Optional[Any] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        is_enabled: bool = True,
    ) -> None:
        """_summary_

        Args:
            name (Optional[str], optional): textual description of the job.
            trigger (Optional[TriggerType], optional): an instance of a trigger class.
            identifier (Optional[str], optional): explicit identifier for the job.
            misfire_grace_time (Optional[int], optional): seconds after the designated runtime that the job is still
                allowed to be run (or ``None`` to allow the job to run no matter how late it is).
            coalesce (Optional[bool], optional): run once instead of many times if the scheduler determines that the
                job should be run more than once in succession.
            max_intances (Optional[int], optional): maximum number of concurrently running instances allowed for this
                job.
            next_run_time (Optional[datetime], optional): when to first run the job, regardless of the trigger (pass
                ``None`` to add the job as paused).
            jobstore (Optional[str], optional): alias of the job store to store the job in.
            executor (Optional[str], optional): alias of the executor to run the job with.
            replace_existing (bool, optional): True to replace an existing job with the same ``id``
                (but retain the number of runs from the existing one).
            args (Optional[Any], optional): list of positional arguments to call func with.
            kwargs (Optional[Dict[str, Any]], optional): dict of keyword arguments to call func with.
            is_enabled (bool, optional): True if the the task to be added to the scheduler.
        """
        self.name = name
        self.trigger = trigger
        self.identifier = identifier
        self.misfire_grace_time = misfire_grace_time or undefined
        self.coalesce = coalesce or undefined
        self.max_intances = max_intances or undefined
        self.next_run_time = next_run_time or undefined
        self.jobstore = jobstore
        self.executor = executor
        self.replace_existing = replace_existing
        self.args = args
        self.kwargs = kwargs
        self.is_enabled = is_enabled
        self.fn = None

    def add_job(self, scheduler: "SchedulerType"):
        try:
            scheduler.add_job(
                func=self.fn,
                trigger=self.trigger,
                args=self.args,
                kwargs=self.kwargs,
                id=self.identifier,
                name=self.name,
                misfire_grace_time=self.misfire_grace_time,
                coalesce=self.coalesce,
                max_instances=self.max_intances,
                next_run_time=self.next_run_time,
                jobstore=self.jobstore,
                executor=self.executor,
                replace_existing=self.replace_existing,
            )
        except Exception as e:
            raise ImproperlyConfigured(str(e))

    def __call__(self, fn: "AnyCallable") -> None:
        """Replaces a function with itself."""
        self.fn = fn
        return self


class scheduler(Task):
    def __init__(
        self,
        *,
        name: Optional[str] = None,
        trigger: Optional[TriggerType] = None,
        identifier: Optional[str] = None,
        misfire_grace_time: Optional[int] = None,
        coalesce: Optional[bool] = None,
        max_intances: Optional[int] = None,
        next_run_time: Optional[datetime] = None,
        jobstore: Optional[str] = "default",
        executor: Optional[str] = "default",
        replace_existing: bool = True,
        extra_args: Optional[Any] = None,
        extra_kwargs: Optional[Dict[str, Any]] = None,
        is_enabled: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            trigger=trigger,
            identifier=identifier,
            misfire_grace_time=misfire_grace_time,
            coalesce=coalesce,
            max_intances=max_intances,
            next_run_time=next_run_time,
            jobstore=jobstore,
            executor=executor,
            replace_existing=replace_existing,
            args=extra_args,
            kwargs=extra_kwargs,
            is_enabled=is_enabled,
        )



import functools
import os
import pkgutil
import sys
from argparse import (
    _AppendConstAction,
    _CountAction,
    _StoreConstAction,
    _SubParsersAction,
)
from collections import defaultdict
from difflib import get_close_matches
from importlib import import_module

import esmerald
from esmerald.core.management.base import (
    BaseDirective,
    DirectiveError,
    DirectiveParser,
    handle_default_options,
)
from esmerald.core.management.color import color_style


def find_directives(management_dir):
    """
    Fiven a path to a management directory, retuens a list of all available directives
    """
    command_dir = os.path.join(management_dir, "directives")
    command_list = [
        name
        for _, name, is_package in pkgutil.iter_modules([command_dir])
        if not is_package and not name.startswith("_")
    ]
    return command_list


def load_directive_class(app_name, name):
    """
    Given a command name and an application name, return the Directive
    class instance. Allow all errors raised by the import process
    (ImportError, AttributeError) to propagate.
    """

    module = import_module("%s.management.directives.%s" % (app_name, name))
    return module.Directive()


@functools.lru_cache(maxsize=None)
def get_directives():
    """
    Return a dictionary mapping command names to their callback applications.

    Look for a management.directives package in esmerald.core. If a directives package exists, register all directives in that package.

    Core directives are always included. If a settings module has been
    specified, also include user-defined directives.

    The dictionary is in the format {command_name: app_name}. Key-value
    pairs from this dictionary can then be used in calls to
    load_directive_class(app_name, command_name)

    If a specific version of a command must be loaded (e.g., with the
    startapp command), the instantiated module can be placed in the
    dictionary in place of the application name.

    The dictionary is cached on the first call and reused on subsequent
    calls.
    """
    command_list = find_directives(__path__[0])
    directives = {name: "esmerald.core" for name in command_list}

    return directives


def call_command(command_name, *args, **options):
    """
    Call the given command, with the given options and args/kwargs.

    This is the primary API you should use for calling specific directives.

    `command_name` may be a string or a command object. Using a string is
    preferred unless the command object is required for further processing or
    testing.

    Some examples:
        call_command('createproject')

        from esmerald.core.management.directives import flush
        cmd = flush.Directive()
        call_command(cmd, verbosity=0, interactive=False)
        # Do something with cmd ...
    """
    if isinstance(command_name, BaseDirective):
        # Directive object passed in.
        command = command_name
        command_name = command.__class__.__module__.split(".")[-1]
    else:
        # Load the command object by name.
        try:
            app_name = get_directives()[command_name]
        except KeyError:
            raise DirectiveError("Unknown command: %r" % command_name)

        if isinstance(app_name, BaseDirective):
            # If the command is already loaded, use it directly.
            command = app_name
        else:
            command = load_directive_class(app_name, command_name)

    # Simulate argument parsing to get the option defaults (see #10080 for details).
    parser = command.create_parser("", command_name)
    # Use the `dest` option name from the parser option
    opt_mapping = {
        min(s_opt.option_strings).lstrip("-").replace("-", "_"): s_opt.dest
        for s_opt in parser._actions
        if s_opt.option_strings
    }
    arg_options = {opt_mapping.get(key, key): value for key, value in options.items()}
    parse_args = []
    for arg in args:
        if isinstance(arg, (list, tuple)):
            parse_args += map(str, arg)
        else:
            parse_args.append(str(arg))

    def get_actions(parser):
        # Parser actions and actions from sub-parser choices.
        for opt in parser._actions:
            if isinstance(opt, _SubParsersAction):
                for sub_opt in opt.choices.values():
                    yield from get_actions(sub_opt)
            else:
                yield opt

    parser_actions = list(get_actions(parser))
    mutually_exclusive_required_options = {
        opt
        for group in parser._mutually_exclusive_groups
        for opt in group._group_actions
        if group.required
    }
    # Any required arguments which are passed in via **options must be passed
    # to parse_args().
    for opt in parser_actions:
        if opt.dest in options and (opt.required or opt in mutually_exclusive_required_options):
            parse_args.append(min(opt.option_strings))
            if isinstance(opt, (_AppendConstAction, _CountAction, _StoreConstAction)):
                continue
            value = arg_options[opt.dest]
            if isinstance(value, (list, tuple)):
                parse_args += map(str, value)
            else:
                parse_args.append(str(value))
    defaults = parser.parse_args(args=parse_args)
    defaults = dict(defaults._get_kwargs(), **arg_options)
    # Raise an error if any unknown options were passed.
    stealth_options = set(command.base_stealth_options + command.stealth_options)
    dest_parameters = {action.dest for action in parser_actions}
    valid_options = (dest_parameters | stealth_options).union(opt_mapping)
    unknown_options = set(options) - valid_options
    if unknown_options:
        raise TypeError(
            "Unknown option(s) for %s command: %s. "
            "Valid options are: %s."
            % (
                command_name,
                ", ".join(sorted(unknown_options)),
                ", ".join(sorted(valid_options)),
            )
        )
    # Move positional args out of options to mimic legacy optparse
    args = defaults.pop("args", ())
    if "skip_checks" not in options:
        defaults["skip_checks"] = True

    return command.execute(*args, **defaults)


class ManagementUtility:
    def __init__(self, argv=None):
        self.argv = argv or sys.argv[:]
        self.prog_name = os.path.basename(self.argv[0])
        if self.prog_name == "__main__.py":
            self.prog_name = "python -m esmerald"
        self.settings_exception = None

    def main_help_text(self, directives_only=False):
        """Return the script's main help text, as a string."""
        if directives_only:
            usage = sorted(get_directives())
        else:
            usage = [
                "",
                "Type '%s help <subdirective>' for help on a specific subdirective."
                % self.prog_name,
                "",
                "Available subdirectives:",
            ]
            directives_dict = defaultdict(lambda: [])
            for name, app in get_directives().items():
                if app == "esmerald.core":
                    app = "esmerald"
                else:
                    app = app.rpartition(".")[-1]
                directives_dict[app].append(name)

            style = color_style()
            for app in sorted(directives_dict):
                usage.append("")
                usage.append(style.NOTICE("[%s]" % app))
                for name in sorted(directives_dict[app]):
                    usage.append("    %s" % name)
            # Output an extra note if settings are not properly configured
            if self.settings_exception is not None:
                usage.append(
                    style.NOTICE(
                        "Note that only Esmerald core directives are listed "
                        "as settings are not properly configured (error: %s)."
                        % self.settings_exception
                    )
                )

        return "\n".join(usage)

    def fetch_command(self, subdirective):
        """
        Try to fetch the given subdirective, printing a message with the
        appropriate command called from the command line (usually
        "esmerald-admin") if it can't be found.
        """
        # Get directives outside of try block to prevent swallowing exceptions
        directives = get_directives()
        try:
            app_name = directives[subdirective]
        except KeyError:
            possible_matches = get_close_matches(subdirective, directives)
            sys.stderr.write("Unknown command: %r" % subdirective)
            if possible_matches:
                sys.stderr.write(". Did you mean %s?" % possible_matches[0])
            sys.stderr.write("\nType '%s help' for usage.\n" % self.prog_name)
            sys.exit(1)
        if isinstance(app_name, BaseDirective):
            # If the command is already loaded, use it directly.
            klass = app_name
        else:
            klass = load_directive_class(app_name, subdirective)
        return klass

    def autocomplete(self):
        """
        Output completion suggestions for BASH.

        The output of this function is passed to BASH's `COMREPLY` variable and
        treated as completion suggestions. `COMREPLY` expects a space
        separated string as the result.

        The `COMP_WORDS` and `COMP_CWORD` BASH environment variables are used
        to get information about the cli input. Please refer to the BASH
        man-page for more information about this variables.

        Subcommand options are saved as pairs. A pair consists of
        the long option string (e.g. '--exclude') and a boolean
        value indicating if the option requires arguments. When printing to
        stdout, an equal sign is appended to options which require arguments.

        Note: If debugging this function, it is recommended to write the debug
        output in a separate file. Otherwise the debug output will be treated
        and formatted as potential completion suggestions.
        """
        # Don't complete if user hasn't sourced bash_completion file.
        if "ESMERALD_AUTO_COMPLETE" not in os.environ:
            return

        cwords = os.environ["COMP_WORDS"].split()[1:]
        cword = int(os.environ["COMP_CWORD"])

        try:
            curr = cwords[cword - 1]
        except IndexError:
            curr = ""

        subdirectives = [*get_directives(), "help"]
        options = [("--help", False)]

        # subdirective
        if cword == 1:
            print(" ".join(sorted(filter(lambda x: x.startswith(curr), subdirectives))))
        # subdirective options
        # special case: the 'help' subdirective has no options
        elif cwords[0] in subdirectives and cwords[0] != "help":
            subcommand_cls = self.fetch_command(cwords[0])
            # special case: add the names of installed apps to options
            parser = subcommand_cls.create_parser("", cwords[0])
            options.extend(
                (min(s_opt.option_strings), s_opt.nargs != 0)
                for s_opt in parser._actions
                if s_opt.option_strings
            )
            # filter out previously specified options from available options
            prev_opts = {x.split("=")[0] for x in cwords[1 : cword - 1]}
            options = (opt for opt in options if opt[0] not in prev_opts)

            # filter options by current input
            options = sorted((k, v) for k, v in options if k.startswith(curr))
            for opt_label, require_arg in options:
                # append '=' to options which require args
                if require_arg:
                    opt_label += "="
                print(opt_label)
        # Exit code of the bash completion function is never passed back to
        # the user, so it's safe to always exit with 0.
        # For more details see #25420.
        sys.exit(0)

    def execute(self):
        """
        Given the command-line arguments, figure out which subdirective is being
        run, create a parser appropriate to that command, and run it.
        """
        try:
            subdirective = self.argv[1]
        except IndexError:
            subdirective = "help"  # Display help if no arguments were given.

        # Preprocess options to extract --settings and --pythonpath.
        # These options could affect the directives that are available, so they
        # must be processed early.
        parser = DirectiveParser(
            prog=self.prog_name,
            usage="%(prog)s subdirective [options] [args]",
            add_help=False,
            allow_abbrev=False,
        )
        parser.add_argument("--pythonpath")
        parser.add_argument("args", nargs="*")  # catch-all
        try:
            options, args = parser.parse_known_args(self.argv[2:])
            handle_default_options(options)
        except DirectiveError:
            pass  # Ignore any option errors at this point.

        self.autocomplete()

        if subdirective == "help":
            if "--directives" in args:
                sys.stdout.write(self.main_help_text(directives_only=True) + "\n")
            elif not options.args:
                sys.stdout.write(self.main_help_text() + "\n")
            else:
                self.fetch_command(options.args[0]).print_help(self.prog_name, options.args[0])
        elif subdirective == "version" or self.argv[1:] == ["--version"]:
            sys.stdout.write(esmerald.__version__ + "\n")
        elif self.argv[1:] in (["--help"], ["-h"]):
            sys.stdout.write(self.main_help_text() + "\n")
        else:
            self.fetch_command(subdirective).run_from_argv(self.argv)


def execute_from_command_line(argv=None):
    """Run a ManagementUtility."""
    utility = ManagementUtility(argv)
    utility.execute()



from inspect import Parameter, Signature
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generator,
    List,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from esmerald.enums import ScopeType
from esmerald.exceptions import (
    ImproperlyConfigured,
    InternalServerError,
    ValidationErrorException,
)
from esmerald.requests import Request
from esmerald.utils.dependency import (
    is_dependency_field,
    should_skip_dependency_validation,
)
from esmerald.utils.helpers import is_optional_union
from esmerald.websockets import WebSocket
from pydantic import BaseConfig, BaseModel, ValidationError, create_model
from pydantic.fields import FieldInfo, Undefined
from pydantic_factories import ModelFactory

if TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict
    from pydantic.typing import AnyCallable
    from starlette.datastructures import URL


UNDEFINED_SENTINELS = {Undefined, Signature.empty}
SKIP_NAMES = {"self", "cls"}
SKIP_VALIDATION_NAMES = {"request", "socket", "state"}


class SignatureModel(BaseModel):
    class Config(BaseConfig):
        arbitrary_types_allowed = True

    dependency_names: ClassVar[Set[str]]
    return_annotation: ClassVar[Any]

    @classmethod
    def parse_values_from_connection_kwargs(
        cls, connection: Union[Request, WebSocket], **kwargs: Any
    ) -> Dict[str, Any]:
        try:
            signature = cls(**kwargs)
            return {key: signature.resolve_field_value(key) for key in cls.__fields__}
        except ValidationError as exc:
            raise cls.construct_exception(connection, exc) from exc

    def resolve_field_value(self, key: str) -> Any:
        value = self.__getattribute__(key)  # pylint: disable=unnecessary-dunder-call
        return value

    @classmethod
    def construct_exception(
        cls, connection: Union[Request, WebSocket], exc: ValidationError
    ) -> Union[InternalServerError, ValidationError]:
        server_errors: List["ErrorDict"] = []
        client_errors: List["ErrorDict"] = []

        for error in exc.errors():
            if cls.is_server_error(error):
                server_errors.append(error)
            else:
                client_errors.append(error)

        method, url = cls.get_connection_method_and_url(connection)

        if client_errors:
            return ValidationErrorException(
                detail=f"Validation failed for {method} {url}", extra=client_errors
            )

        return InternalServerError(
            detail=f"A dependency failed validation for {method} {url}",
            extra=server_errors,
        )

    @classmethod
    def is_server_error(cls, error: "ErrorDict") -> bool:
        """Check whether given validation error is a server error."""
        return error["loc"][-1] in cls.dependency_names

    @staticmethod
    def get_connection_method_and_url(connection: Union[Request, WebSocket]) -> Tuple[str, "URL"]:
        """Extract method and URL from Request or WebSocket."""
        method = ScopeType.WEBSOCKET if isinstance(connection, WebSocket) else connection.method
        return method, connection.url


class SignatureParameter:
    __slots__ = (
        "annotation",
        "default",
        "name",
        "optional",
    )

    annotation: Any
    default: Any
    name: str
    optional: bool

    def __init__(self, fn_name: str, parameter_name: str, parameter: Parameter) -> None:
        if parameter.annotation is Signature.empty:
            raise ImproperlyConfigured(
                f"Kwarg {parameter_name} of {fn_name} does not have a type annotation. If it "
                f"should receive any value, use the 'Any' type."
            )
        self.annotation = parameter.annotation
        self.default = parameter.default
        self.name = parameter_name
        self.optional = is_optional_union(parameter.annotation)

    @property
    def default_defined(self) -> bool:
        return self.default not in UNDEFINED_SENTINELS


class SignatureModelFactory:
    __slots__ = (
        "defaults",
        "dependency_names",
        "field_definitions",
        "fn_name",
        "signature",
    )

    def __init__(
        self,
        fn: "AnyCallable",
        dependency_names: Set[str],
    ) -> None:
        if fn is None:
            raise ImproperlyConfigured(
                "Parameter `fn` to `SignatureModelFactory` cannot be `None`."
            )
        self.signature = Signature.from_callable(fn)
        self.fn_name = fn.__name__ if hasattr(fn, "__name__") else "anonymous"
        self.field_definitions: Dict[str, Any] = {}
        self.defaults: Dict[str, Any] = {}
        self.dependency_names = dependency_names

    def check_for_unprovided_dependency(self, parameter: SignatureParameter) -> None:
        if parameter.optional:
            return
        if not is_dependency_field(parameter.default):
            return
        field_info: FieldInfo = parameter.default
        if field_info.default is not Undefined:
            return
        if parameter.name not in self.dependency_names:
            raise ImproperlyConfigured(
                f"Explicit dependency '{parameter.name}' for '{self.fn_name}' has no default value, "
                f"or provided dependency."
            )

    def collect_dependency_names(self, parameter: SignatureParameter) -> None:
        if is_dependency_field(parameter.default):
            self.dependency_names.add(parameter.name)

    def set_field_default(self, parameter: SignatureParameter) -> None:
        if parameter.default_defined:
            self.defaults[parameter.name] = parameter.default

    @staticmethod
    def create_field_definition_from_parameter(
        parameter: SignatureParameter,
    ) -> Tuple[Any, Any]:
        if parameter.default_defined:
            field_definition = (parameter.annotation, parameter.default)
        elif not parameter.optional:
            field_definition = (parameter.annotation, ...)
        else:
            field_definition = (parameter.annotation, None)
        return field_definition

    @property
    def signature_parameters(self) -> Generator[SignatureParameter, None, None]:
        for name, parameter in self.signature.parameters.items():
            if name in SKIP_NAMES:
                continue
            yield SignatureParameter(self.fn_name, name, parameter)

    def should_skip_parameter_validation(self, parameter: SignatureParameter) -> bool:
        return parameter.name in SKIP_VALIDATION_NAMES or should_skip_dependency_validation(
            parameter.default
        )

    def create_signature_model(self) -> Type[SignatureModel]:
        try:
            for parameter in self.signature_parameters:
                self.check_for_unprovided_dependency(parameter)
                self.collect_dependency_names(parameter)
                self.set_field_default(parameter)
                if self.should_skip_parameter_validation(parameter):
                    # pydantic has issues with none-pydantic classes that receive generics
                    self.field_definitions[parameter.name] = (Any, ...)
                    continue
                if ModelFactory.is_constrained_field(parameter.default):
                    self.field_definitions[parameter.name] = (parameter.default, ...)
                    continue
                self.field_definitions[
                    parameter.name
                ] = self.create_field_definition_from_parameter(parameter)
            model: Type[SignatureModel] = create_model(
                self.fn_name + "_signature_model",
                __base__=SignatureModel,
                **self.field_definitions,
            )
            model.return_annotation = self.signature.return_annotation
            model.dependency_names = self.dependency_names
            return model
        except TypeError as e:
            raise ImproperlyConfigured(
                f"Error creating signature model for '{self.fn_name}': '{e}'"
            ) from e


def get_signature_model(value: Any) -> Type[SignatureModel]:
    try:
        return cast("Type[SignatureModel]", getattr(value, "signature_model"))
    except AttributeError as e:  # pragma: no cover
        raise ImproperlyConfigured(
            f"The 'signature_model' attribute for {value} is not set"
        ) from e



from typing import Any, Dict, Optional

color_names = ("black", "red", "green", "yellow", "blue", "magenta", "cyan", "white")
foreground = {color_names[x]: "3%s" % x for x in range(8)}
background = {color_names[x]: "4%s" % x for x in range(8)}

RESET = "0"
opt_dict = {
    "bold": "1",
    "underscore": "4",
    "blink": "5",
    "reverse": "7",
    "conceal": "8",
}


def colorize(text: Optional[str] = "", opts: Optional[Any] = None, **kwargs: Dict[str, Any]):
    """
    Return your text, enclosed in ANSI graphics codes.

    Depends on the keyword arguments 'fg' and 'bg', and the contents of
    the opts tuple/list.

    Return the RESET code if no parameters are given.

    Valid colors:
        'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'

    Valid options:
        'bold'
        'underscore'
        'blink'
        'reverse'
        'conceal'
        'noreset' - string will not be auto-terminated with the RESET code

    Examples:
        colorize('hello', fg='red', bg='blue', opts=('blink',))
        colorize()
        colorize('goodbye', opts=('underscore',))
        print(colorize('first line', fg='red', opts=('noreset',)))
        print('this should be red too')
        print(colorize('and so should this'))
        print('this should not be red')
    """
    if not opts:
        opts = ()

    code_list = []
    if text == "" and len(opts) == 1 and opts[0] == "reset":
        return "\x1b[%sm" % RESET
    for k, v in kwargs.items():
        if k == "fg":
            code_list.append(foreground[v])
        elif k == "bg":
            code_list.append(background[v])
    for o in opts:
        if o in opt_dict:
            code_list.append(opt_dict[o])
    if "noreset" not in opts:
        text = "%s\x1b[%sm" % (text or "", RESET)
    return "%s%s" % (("\x1b[%sm" % ";".join(code_list)), text or "")


def make_style(opts: Optional[Any] = None, **kwargs: Dict[str, Any]):
    """
    Return a function with default parameters for colorize()

    Example:
        bold_red = make_style(opts=('bold',), fg='red')
        print(bold_red('hello'))
        KEYWORD = make_style(fg='yellow')
        COMMENT = make_style(fg='blue', opts=('bold',))
    """
    if not opts:
        opts = ()
    return lambda text: colorize(text, opts, **kwargs)


NOCOLOR_PALETTE = "nocolor"
DARK_PALETTE = "dark"
LIGHT_PALETTE = "light"

PALETTES = {
    NOCOLOR_PALETTE: {
        "ERROR": {},
        "SUCCESS": {},
        "WARNING": {},
        "NOTICE": {},
        "SQL_FIELD": {},
        "SQL_COLTYPE": {},
        "SQL_KEYWORD": {},
        "SQL_TABLE": {},
        "HTTP_INFO": {},
        "HTTP_SUCCESS": {},
        "HTTP_REDIRECT": {},
        "HTTP_NOT_MODIFIED": {},
        "HTTP_BAD_REQUEST": {},
        "HTTP_NOT_FOUND": {},
        "HTTP_SERVER_ERROR": {},
        "MIGRATE_HEADING": {},
        "MIGRATE_LABEL": {},
    },
    DARK_PALETTE: {
        "ERROR": {"fg": "red", "opts": ("bold",)},
        "SUCCESS": {"fg": "green", "opts": ("bold",)},
        "WARNING": {"fg": "yellow", "opts": ("bold",)},
        "NOTICE": {"fg": "red"},
        "SQL_FIELD": {"fg": "green", "opts": ("bold",)},
        "SQL_COLTYPE": {"fg": "green"},
        "SQL_KEYWORD": {"fg": "yellow"},
        "SQL_TABLE": {"opts": ("bold",)},
        "HTTP_INFO": {"opts": ("bold",)},
        "HTTP_SUCCESS": {},
        "HTTP_REDIRECT": {"fg": "green"},
        "HTTP_NOT_MODIFIED": {"fg": "cyan"},
        "HTTP_BAD_REQUEST": {"fg": "red", "opts": ("bold",)},
        "HTTP_NOT_FOUND": {"fg": "yellow"},
        "HTTP_SERVER_ERROR": {"fg": "magenta", "opts": ("bold",)},
        "MIGRATE_HEADING": {"fg": "cyan", "opts": ("bold",)},
        "MIGRATE_LABEL": {"opts": ("bold",)},
    },
    LIGHT_PALETTE: {
        "ERROR": {"fg": "red", "opts": ("bold",)},
        "SUCCESS": {"fg": "green", "opts": ("bold",)},
        "WARNING": {"fg": "yellow", "opts": ("bold",)},
        "NOTICE": {"fg": "red"},
        "SQL_FIELD": {"fg": "green", "opts": ("bold",)},
        "SQL_COLTYPE": {"fg": "green"},
        "SQL_KEYWORD": {"fg": "blue"},
        "SQL_TABLE": {"opts": ("bold",)},
        "HTTP_INFO": {"opts": ("bold",)},
        "HTTP_SUCCESS": {},
        "HTTP_REDIRECT": {"fg": "green", "opts": ("bold",)},
        "HTTP_NOT_MODIFIED": {"fg": "green"},
        "HTTP_BAD_REQUEST": {"fg": "red", "opts": ("bold",)},
        "HTTP_NOT_FOUND": {"fg": "red"},
        "HTTP_SERVER_ERROR": {"fg": "magenta", "opts": ("bold",)},
        "MIGRATE_HEADING": {"fg": "cyan", "opts": ("bold",)},
        "MIGRATE_LABEL": {"opts": ("bold",)},
    },
}
DEFAULT_PALETTE = DARK_PALETTE


def parse_color_setting(config_string: str):
    """Parse a ESMERALD_COLORS environment variable to produce the system palette

    The general form of a palette definition is:

        "palette;role=fg;role=fg/bg;role=fg,option,option;role=fg/bg,option,option"

    where:
        palette is a named palette; one of 'light', 'dark', or 'nocolor'.
        role is a named style used by Esmerald
        fg is a foreground color.
        bg is a background color.
        option is a display options.

    Specifying a named palette is the same as manually specifying the individual
    definitions for each role. Any individual definitions following the palette
    definition will augment the base palette definition.

    Valid roles:
        'error', 'success', 'warning', 'notice', 'sql_field', 'sql_coltype',
        'sql_keyword', 'sql_table', 'http_info', 'http_success',
        'http_redirect', 'http_not_modified', 'http_bad_request',
        'http_not_found', 'http_server_error', 'migrate_heading',
        'migrate_label'

    Valid colors:
        'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'

    Valid options:
        'bold', 'underscore', 'blink', 'reverse', 'conceal', 'noreset'
    """
    if not config_string:
        return PALETTES[DEFAULT_PALETTE]

    # Split the color configuration into parts
    parts = config_string.lower().split(";")
    palette = PALETTES[NOCOLOR_PALETTE].copy()
    for part in parts:
        if part in PALETTES:
            # A default palette has been specified
            palette.update(PALETTES[part])
        elif "=" in part:
            # Process a palette defining string
            definition = {}

            # Break the definition into the role,
            # plus the list of specific instructions.
            # The role must be in upper case
            role, instructions = part.split("=")
            role = role.upper()

            styles = instructions.split(",")
            styles.reverse()

            # The first instruction can contain a slash
            # to break apart fg/bg.
            colors = styles.pop().split("/")
            colors.reverse()
            fg = colors.pop()
            if fg in color_names:
                definition["fg"] = fg
            if colors and colors[-1] in color_names:
                definition["bg"] = colors[-1]

            # All remaining instructions are options
            opts = tuple(s for s in styles if s in opt_dict)
            if opts:
                definition["opts"] = opts

            # The nocolor palette has all available roles.
            # Use that palette as the basis for determining
            # if the role is valid.
            if role in PALETTES[NOCOLOR_PALETTE] and definition:
                palette[role] = definition

    # If there are no colors specified, return the empty palette.
    if palette == PALETTES[NOCOLOR_PALETTE]:
        return None
    return palette



from typing import Any

from esmerald.utils.constants import EXTRA_KEY_IS_DEPENDENCY, EXTRA_KEY_SKIP_VALIDATION
from pydantic.fields import FieldInfo


def is_dependency_field(val: Any) -> bool:
    return isinstance(val, FieldInfo) and bool(val.extra.get(EXTRA_KEY_IS_DEPENDENCY))


def should_skip_dependency_validation(val: Any) -> bool:
    return is_dependency_field(val) and val.extra.get(EXTRA_KEY_SKIP_VALIDATION)



import time
from typing import Any, Dict, Optional

from apscheduler.schedulers.asyncio import AsyncIOScheduler as APSAsyncIOScheduler

# ┌───────────── minute (0 - 59)
# │ ┌───────────── hour (0 - 23)
# │ │ ┌───────────── day of the month (1 - 31)
# │ │ │ ┌───────────── month (1 - 12)
# │ │ │ │ ┌───────────── day of the week (0 - 6) (Sunday to Saturday;
# │ │ │ │ │                                   7 is also Sunday on some systems)
# │ │ │ │ │
# │ │ │ │ │
# * * * * * <command to execute>


def get_time():
    return time.strftime("%A, %d. %B %Y %I:%M:%S %p")


class AsyncIOScheduler(APSAsyncIOScheduler):
    def __init__(self, gconfig: Optional[Dict[Any, Any]] = None, **options):
        if not gconfig:
            gconfig = {}
        super().__init__(gconfig, **options)



"""
Base classes for writing management directives (named directives which can
be executed through `esmerald-admin`).
"""
import argparse
import os
import sys
from argparse import ArgumentParser, HelpFormatter
from io import TextIOBase
from typing import Any, Dict, Optional

import esmerald
from esmerald.core.management.color import color_style, no_style


class DirectiveError(Exception):
    """
    Exception class indicating a problem while executing a directive.
    """

    def __init__(self, *args: Any, returncode: int = 1, **kwargs: Dict[str, Any]):
        self.returncode = returncode
        super().__init__(*args, **kwargs)


class SystemCheckError(DirectiveError):
    """
    The system check framework detected unrecoverable errors.
    """

    ...


class DirectiveParser(ArgumentParser):
    """
    Customized ArgumentParser class to improve some error messages and prevent SystemExit.
    """

    def __init__(
        self,
        *,
        missing_args_message: Optional[str] = None,
        called_from_command_line: Optional[Any] = None,
        **kwargs: Dict[str, Any]
    ):
        self.missing_args_message = missing_args_message
        self.called_from_command_line = called_from_command_line
        super().__init__(**kwargs)

    def parse_args(self, args: Any = None, namespace: str = None):
        if self.missing_args_message and not (
            args or any(not arg.startswith("-") for arg in args)
        ):
            self.error(self.missing_args_message)
        return super().parse_args(args, namespace)

    def error(self, message):
        if self.called_from_command_line:
            super().error(message)
        else:
            raise DirectiveError("Error: %s" % message)


def handle_default_options(options: Any):
    """
    Include any default options that all directives should accept here
    so that ManagementUtility can handle them before searching for
    user directives.
    """
    if options.pythonpath:
        sys.path.insert(0, options.pythonpath)


class EsmeraldHelpFormatter(HelpFormatter):
    """
    Customized formatter so that command-specific arguments appear in the
    --help output before arguments common to all directives.
    """

    show_last = {
        "--version",
        "--verbosity",
        "--traceback",
        "--settings",
        "--pythonpath",
        "--no-color",
        "--force-color",
    }

    def _reordered_actions(self, actions: Any):
        return sorted(actions, key=lambda a: set(a.option_strings) & self.show_last != set())

    def add_usage(self, usage: str, actions: Any, *args: Any, **kwargs: Dict[str, Any]):
        super().add_usage(usage, self._reordered_actions(actions), *args, **kwargs)

    def add_arguments(self, actions: Any):
        super().add_arguments(self._reordered_actions(actions))


class OutputWrapper(TextIOBase):
    @property
    def style_func(self):
        return self._style_func

    @style_func.setter
    def style_func(self, style_func: Any):
        if style_func and self.isatty():
            self._style_func = style_func
        else:
            self._style_func = lambda x: x

    def __init__(self, out: Any, ending="\n"):
        self._out = out
        self.style_func = None
        self.ending = ending

    def __getattr__(self, name: str):
        return getattr(self._out, name)

    def flush(self):
        if hasattr(self._out, "flush"):
            self._out.flush()

    def isatty(self):
        return hasattr(self._out, "isatty") and self._out.isatty()

    def write(
        self,
        msg: str = "",
        style_func: Optional[Any] = None,
        ending: Optional[Any] = None,
    ):
        ending = self.ending if ending is None else ending
        if ending and not msg.endswith(ending):
            msg += ending
        style_func = style_func or self.style_func
        self._out.write(style_func(msg))


class BaseDirective:
    """
    The base class from which all management directives ultimately
    derive.

    Use this class if you want access to all of the mechanisms which
    parse the command-line arguments and work out what code to call in
    response; if you don't need to change any of that behavior,
    consider using one of the subclasses defined in this file.

    If you are interested in overriding/customizing various aspects of
    the command-parsing and -execution behavior, the normal flow works
    as follows:

    1. ``esmerald-admin`` loads the command class and calls its ``run_from_argv()`` method.

    2. The ``run_from_argv()`` method calls ``create_parser()`` to get
       an ``ArgumentParser`` for the arguments, parses them, performs
       any environment changes requested by options like
       ``pythonpath``, and then calls the ``execute()`` method,
       passing the parsed arguments.

    3. The ``execute()`` method attempts to carry out the command by
       calling the ``handle()`` method with the parsed arguments; any
       output produced by ``handle()`` will be printed to standard
       output and, if the command is intended to produce a block of
       SQL statements, will be wrapped in ``BEGIN`` and ``COMMIT``.

    4. If ``handle()`` or ``execute()`` raised any exception (e.g.
       ``DirectiveError``), ``run_from_argv()`` will  instead print an error
       message to ``stderr``.

    Thus, the ``handle()`` method is typically the starting point for
    subclasses; many built-in directives and command types either place
    all of their logic in ``handle()``, or perform some additional
    parsing work in ``handle()`` and then delegate from it to more
    specialized methods as needed.

    Several attributes affect behavior at various steps along the way:

    ``help``
        A short description of the command, which will be printed in
        help messages.

    ``output_transaction``
        A boolean indicating whether the command outputs SQL
        statements; if ``True``, the output will automatically be
        wrapped with ``BEGIN;`` and ``COMMIT;``. Default value is
        ``False``.

    ``requires_migrations_checks``
        A boolean; if ``True``, the command prints a warning if the set of
        migrations on disk don't match the migrations in the database.

    ``requires_system_checks``
        A list or tuple of tags, e.g. [Tags.staticfiles, Tags.models]. System
        checks registered in the chosen tags will be checked for errors prior
        to executing the command. The value '__all__' can be used to specify
        that all system checks should be performed. Default value is '__all__'.

        To validate an individual application's models
        rather than all applications' models, call
        ``self.check(app_configs)`` from ``handle()``, where ``app_configs``
        is the list of application's configuration provided by the
        app registry.

    ``stealth_options``
        A tuple of any options the command uses which aren't defined by the
        argument parser.
    """

    # Metadata about this command.
    help = ""

    _called_from_command_line = False
    output_transaction = False  # Whether to wrap the output in a "BEGIN; COMMIT;"
    base_stealth_options = ("stderr", "stdout")
    stealth_options = ()
    suppressed_base_arguments = set()

    def __init__(self, stdout=None, stderr=None, no_color=False, force_color=False):
        self.stdout = OutputWrapper(stdout or sys.stdout)
        self.stderr = OutputWrapper(stderr or sys.stderr)
        if no_color and force_color:
            raise DirectiveError("'no_color' and 'force_color' can't be used together.")
        if no_color:
            self.style = no_style()
        else:
            self.style = color_style(force_color)
            self.stderr.style_func = self.style.ERROR

    def get_version(self):
        """
        Return the Esmerald version, which should be correct for all built-in
        Esmerald directives. User-supplied directives can override this method to
        return their own version.
        """
        return esmerald.__version__

    def create_parser(self, prog_name: Any, subdirective: Any, **kwargs: Dict[str, Any]):
        """
        Create and return the ``ArgumentParser`` which will be used to
        parse the arguments to this command.
        """
        kwargs.setdefault("formatter_class", EsmeraldHelpFormatter)
        parser = DirectiveParser(
            prog="%s %s" % (os.path.basename(prog_name), subdirective),
            description=self.help or None,
            missing_args_message=getattr(self, "missing_args_message", None),
            called_from_command_line=getattr(self, "_called_from_command_line", None),
            **kwargs,
        )
        self.add_base_argument(
            parser,
            "--version",
            action="version",
            version=self.get_version(),
            help="Show program's version number and exit.",
        )
        self.add_base_argument(
            parser,
            "-v",
            "--verbosity",
            default=1,
            type=int,
            choices=[0, 1, 2, 3],
            help=(
                "Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, "
                "3=very verbose output"
            ),
        )
        self.add_base_argument(
            parser,
            "--settings",
            help=(
                "The Python path to a settings module, e.g. "
                '"myproject.settings.main". If this isn\'t provided, the '
                "ESMERALD_SETTINGS_MODULE environment variable will be used."
            ),
        )
        self.add_base_argument(
            parser,
            "--pythonpath",
            help=(
                "A directory to add to the Python path, e.g. "
                '"/home/esmeraldprojects/myproject".'
            ),
        )
        self.add_base_argument(
            parser,
            "--traceback",
            action="store_true",
            help="Raise on DirectiveError exceptions.",
        )
        self.add_base_argument(
            parser,
            "--no-color",
            action="store_true",
            help="Don't colorize the command output.",
        )
        self.add_base_argument(
            parser,
            "--force-color",
            action="store_true",
            help="Force colorization of the command output.",
        )
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser: Any):
        """
        Entry point for subclassed directives to add custom arguments.
        """

    def add_base_argument(self, parser: Any, *args: Any, **kwargs: Dict[str, Any]):
        """
        Call the parser's add_argument() method, suppressing the help text
        according to BaseDirective.suppressed_base_arguments.
        """
        for arg in args:
            if arg in self.suppressed_base_arguments:
                kwargs["help"] = argparse.SUPPRESS
                break
        parser.add_argument(*args, **kwargs)

    def print_help(self, prog_name, subdirective):
        """
        Print the help message for this command, derived from
        ``self.usage()``.
        """
        parser = self.create_parser(prog_name, subdirective)
        parser.print_help()

    def run_from_argv(self, argv: Any):
        """
        Set up any environment changes requested (e.g., Python path
        and Esmerald settings), then run this command. If the
        command raises a ``DirectiveError``, intercept it and print it sensibly
        to stderr. If the ``--traceback`` option is present or the raised
        ``Exception`` is not ``DirectiveError``, raise it.
        """
        self._called_from_command_line = True
        parser = self.create_parser(argv[0], argv[1])

        options = parser.parse_args(argv[2:])
        cmd_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop("args", ())
        handle_default_options(options)
        try:
            self.execute(*args, **cmd_options)
        except DirectiveError as e:
            if options.traceback:
                raise

            # SystemCheckError takes care of its own formatting.
            if isinstance(e, SystemCheckError):
                self.stderr.write(str(e), lambda x: x)
            else:
                self.stderr.write("%s: %s" % (e.__class__.__name__, e))
            sys.exit(e.returncode)

    def execute(self, *args: Any, **options: Dict[str, Any]):
        """
        Try to execute this command, performing system checks if needed (as
        controlled by the ``requires_system_checks`` attribute, except if
        force-skipped).
        """
        if options["force_color"] and options["no_color"]:
            raise DirectiveError(
                "The --no-color and --force-color options can't be used together."
            )
        if options["force_color"]:
            self.style = color_style(force_color=True)
        elif options["no_color"]:
            self.style = no_style()
            self.stderr.style_func = None
        if options.get("stdout"):
            self.stdout = OutputWrapper(options["stdout"])
        if options.get("stderr"):
            self.stderr = OutputWrapper(options["stderr"])

        output = self.handle(*args, **options)
        if output:
            self.stdout.write(output)
        return output

    def handle(self, *args: Any, **options: Dict[str, Any]):
        """
        The actual logic of the command. Subclasses must implement
        this method.
        """
        raise NotImplementedError("subclasses of BaseDirective must provide a handle() method")



from apscheduler.triggers.combining import AndTrigger, OrTrigger
from apscheduler.triggers.cron import BaseTrigger, CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger

__all__ = [
    "AndTrigger",
    "OrTrigger",
    "BaseTrigger",
    "CronTrigger",
    "DateTrigger",
    "IntervalTrigger",
]
